﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Xml.Serialization;

namespace Q_Learning
{
	using Pair = KeyValuePair<int, bool>;
	public partial class MainForm : Form
	{
		AI ai = new AI();
		Dictionary<Status, double> Q;
		const string file = "data.txt";
		public MainForm()
		{
			InitializeComponent();
			Q = new Dictionary<Status, double>();
		}
		private void MainForm_Load(object sender, EventArgs e)
		{
			Directory.SetCurrentDirectory(Path.GetDirectoryName(GetType().Assembly.Location));
			XmlSerializer xs = new XmlSerializer(typeof(List<string>));
			try
			{
				using (var sr = new StreamReader(file, Encoding.GetEncoding(932)))
				{
					List<string> lst = (List<string>)xs.Deserialize(sr);
					foreach (string s in lst)
					{
						string[] ss = s.Split(',');
						Q[Status.Create(int.Parse(ss[0]))] = double.Parse(ss[1]);
					}
				}
				MakeList();
			}
			catch (Exception) { }
		}
		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (!cbSave.Checked) return;
			List<string> lst = new List<string>();
			foreach (KeyValuePair<Status, double> kv in Q)
			{
				lst.Add(string.Format("{0},{1}", kv.Key.Value, Math.Round(kv.Value,6)));
			}
			XmlSerializer xs = new XmlSerializer(typeof(List<string>));
			using (var sw = new StreamWriter(file, false, Encoding.GetEncoding(932)))
				xs.Serialize(sw, lst);
		}
		private void btnInit_Click(object sender, EventArgs e)
		{
			Q.Clear();
			ban1.Status.Reset();
			Refresh();
			MakeList();
		}
		private void btnDo_Click(object sender, EventArgs e)
		{
			Cursor = Cursors.WaitCursor;
			double eps = 0.01, alpha = 0.1, gamma = 0.999999;
			int n = 10000, kgn = 1000, kgc = 0, kgw = 0, kgl = 0;
			for (int iter = 0; iter < n; ++iter)
			{
				Status s = new Status();
				while (true)
				{
					// Q学習
					double qsa = double.MinValue;
					int act1 = -1;
					List<int> lst = s.Actions(EStatus.First);
					if (ai.rand.NextDouble() < eps)
					{
						act1 = lst[ai.rand.Next(lst.Count)];
					}
					else
					{
						double d;
						foreach (int i in lst)
						{
							bool bFound = Q.TryGetValue(s.Next(i, EStatus.First), out d);
							if (qsa < d || (qsa <= d && !bFound))
							{
								qsa = d;
								act1 = i;
							}
						}
						if (act1 < 0)
							act1 = lst[ai.rand.Next(lst.Count)];
					}
					s[act1] = EStatus.First;
					Status pre = s.Normalize();
					if (qsa == double.MinValue)
						Q.TryGetValue(pre, out qsa);
					int res = AI.Check(s); // 勝敗
					// 相手
					Pair pr = new KeyValuePair<int,bool>();
					if (res == 0)
					{
						pr = ai.NextAction(s, false);
						s[-pr.Key - 1] = EStatus.Second;
						if (pr.Value) res = 2;
					}
					double reword = res == 1 ? 1 : (res == 2 ? -1 : 0);
					if (res != 0) // 勝負がついた
						qsa = reword;
					else
						qsa += alpha * (reword + gamma * Max(Q, s) - qsa);
					Q[pre] = qsa;
					if (res > 0)
					{
						if (res == 1) ++kgw;
						else if (pr.Value) ++kgl;
						break;
					}
				}
				if (++kgc == kgn)
				{
					Console.WriteLine("勝率 {0}  敗率 {1}",
						kgw / (double)kgn,
						kgl / (double)kgn);
					kgc = kgw = kgl = 0;
				}
			}
			MakeList();
			Cursor = Cursors.Default;
		}
		public double Max(Dictionary<Status, double> Q, Status s)
		{
			double d, max = double.MinValue;
			for (int i = 0; i < 9; ++i)
			{
				if (s[i] != EStatus.None) continue;
				if (Q.TryGetValue(s.Next(i, EStatus.First), out d) && max < d) max = d;
			}
			if (max == double.MinValue) return 0;
			return max;
		}
		private void MakeList()
		{
			listBox1.Items.Clear();
			List<KeyValuePair<Status, double>> l = Q.ToList();
			l.Sort(delegate(KeyValuePair<Status, double> kv1, KeyValuePair<Status, double> kv2)
				{ return kv1.Key.CompareTo(kv2.Key); });
			int i = 0;
			foreach (KeyValuePair<Status, double> kv in l)
			{
				listBox1.Items.Add(new KeyValuePair<int, KeyValuePair<Status, double>>(++i, kv));
			}
		}
		private void listBox1_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (listBox1.SelectedIndex < 0) return;
			var kv = (KeyValuePair<int, KeyValuePair<Status, double>>)listBox1.SelectedItem;
			ban1.Status = kv.Value.Key;
			Refresh();
		}
	}
}
