﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Q_Learning
{
	public enum EStatus
	{
		None = 0,
		First,
		Second,
		Error
	}
	[System.Diagnostics.DebuggerVisualizer(typeof(StatusVisualizer))]
	[Serializable]
	public struct Status : IComparable<Status>, ICloneable
	{
		public int Value;
		public static Status Create(int i)
		{
			Status s = new Status();
			s.Value = i;
			return s;
		}
		public void Reset()
		{
			Value = 0;
		}
		public EStatus this[int i]
		{
			get
			{
				return (EStatus)((Value >> (2 * i)) & 3);
			}
			set
			{
				Value &= ~(3 << (2 * i));
				Value |= (int)value << (2 * i);
			}
		}
		public List<int> Actions(EStatus es)
		{
			List<int> lst = new List<int>();
			HashSet<int> h = new HashSet<int>();
			for (int i = 0; i < 9; ++i)
			{
				Status s = this;
				if (s[i] != EStatus.None) continue;
				s[i] = es;
				bool bFound = false;
				for (int j = 0; j < 8; ++j)
					if (bFound = h.Contains(s.Transform(j).Value)) break;
				if (bFound) continue;
				h.Add(s.Value);
				lst.Add(i);
			}
			return lst;
		}
		public Status Next(int i, EStatus es)
		{
			Status s = this;
			s[i] = es;
			return s.Normalize();
		}
		public Status Normalize()
		{
			Status st = this;
			return Enumerable.Range(0, 8).Select(i => st.Transform(i)).Min();
		}
		private static int[][] TransformIndex =
		{
			 new [] {0,2,3,5,6,8},
			 new [] {0,6,1,7,2,8},
			 new [] {1,3,2,6,5,7},
			 new [] {1,5,0,8,3,7},
			 new [] {0,8,1,7,2,6,3,5},
			 new [] {0,6,2,8,2,6,1,3,5,7,3,5},
			 new [] {0,2,6,8,2,6,1,3,5,7,1,7},
		};
		public Status Transform(int k)
		{
			k %= 8;
			Status s = this;
			if (k > 0)
			{
				int[] ti = TransformIndex[k - 1];
				for (int i = 0; i < ti.Length / 2; ++i) s.Swap(ti[2 * i], ti[2 * i + 1]);
			}
			return s;
		}
		private void Swap(int i, int j)
		{
			int k = 2 * (i - j);
			if (k == 0) return;
			int vi = Value & (3 << (2 * i));
			int vj = Value & (3 << (2 * j));
			Value -= vi + vj;
			if (k > 0)
			{
				Value |= vi >> k;
				Value |= vj << k;
			}
			else
			{
				Value |= vi << -k;
				Value |= vj >> -k;
			}
		}
		public int CompareTo(Status s)
		{
			return Value.CompareTo(s.Value);
		}
		public object Clone()
		{
			return MemberwiseClone();
		}
		public override int GetHashCode()
		{
			return Value;
		}
		public override bool Equals(object o)
		{
			Status s = (Status)o;
			return Value == s.Value;
		}
		public override string ToString()
		{
			return Value.ToString();
		}
	}
}
